#![allow(unused_parens)]

use crate::posi;
use crate::chirp;
use crate::gaussian;

struct SignalInput
{
	template_rise: Vec<f32>,
	template_fall: Vec<f32>,
	norm_rise: f32,
	norm_fall: f32,
	symbol_length: f32,
	buffer: posi::Buffer<f32>,
	sum: f32,
}

impl SignalInput
{
	fn new
	(
		center_freq: f32,
		deviation: f32,
		symbol_length: f32,
	) -> 
		SignalInput
	{
		let upper_freq = (center_freq + deviation).max(0.0).min(1.0);
		let lower_freq = (center_freq - deviation).max(0.0).min(1.0);

		// these have to be reversed, since the input buffer shifts in at index 0 and is thus a mirror image
		let template_rise:Vec<f32> = chirp::generate(upper_freq, lower_freq, symbol_length, 0.125);
		let template_fall:Vec<f32> = chirp::generate(lower_freq, upper_freq, symbol_length, 0.125);
		
		let norm_rise = template_rise.iter().sum();
		let norm_fall = template_fall.iter().sum();

		SignalInput
		{
			template_rise: template_rise,
			template_fall: template_fall,
			norm_rise: norm_rise,
			norm_fall: norm_fall,
			symbol_length: symbol_length,
			buffer: posi::Buffer::new(symbol_length.round() as usize),
			sum: 0.0,
		}
	}

	fn ingest
	(
		&mut self,
		sample: f32,
	) {
		let new_value = sample;
		let old_value = self.buffer.last();

		self.sum -= old_value.abs();
		self.sum += new_value.abs();

		self.buffer.shift_in(new_value);
	}

	fn process
	(
		&self,
	) -> 
		(f32, f32)
	{
		let local_signal = self.buffer.to_vec();

		let mut rise_value:f32 = 0.0;
		let mut fall_value:f32 = 0.0;

		if self.sum > 0.0
		{
			for (&a, &b) in 
			(
				local_signal.iter()
			).zip(
				self.template_rise.iter()
			) {
				rise_value += a * b;
			}

			for (&a, &b) in 
			(
				local_signal.iter()
			).zip(
				self.template_fall.iter()
			) {
				fall_value += a * b;
			}

			rise_value /= self.norm_rise * self.sum / self.symbol_length;
			fall_value /= self.norm_fall * self.sum / self.symbol_length;
		}

		return (rise_value, fall_value);
	}
}

struct CorrelFilter
{
	kernel: Vec<f32>,
	buffer_rise: posi::Buffer<f32>,
	buffer_fall: posi::Buffer<f32>,
}

impl CorrelFilter
{
	fn new
	(
		freq: f32,
	) -> 
		CorrelFilter
	{
		let kernel = gaussian::generate(freq);
		let size = kernel.len();

		CorrelFilter
		{
			kernel: kernel,
			buffer_rise: posi::Buffer::new(size),
			buffer_fall: posi::Buffer::new(size),
		}
	}

	fn ingest
	(
		&mut self,
		rise: f32,
		fall: f32,
	) {
		self.buffer_rise.shift_in(rise);
		self.buffer_fall.shift_in(fall);
	}

	fn lo_process
	(
		&self,
	) ->
		(f32, f32)
	{
		let local_rise = self.buffer_rise.to_vec();
		let local_fall = self.buffer_fall.to_vec();

		let mut rise_result:f32 = 0.0;
		let mut fall_result:f32 = 0.0;

		for (&a, &b) in
		(
			local_rise.iter()
		).zip(
			self.kernel.iter()
		) {
			rise_result += a * b;
		}

		for (&a, &b) in
		(
			local_fall.iter()
		).zip(
			self.kernel.iter()
		) {
			fall_result += a * b;
		}

		return (rise_result, fall_result);
	}

	fn hi_process
	(
		&self,
	) ->
		(f32, f32)
	{
		let (rise_result, fall_result) = self.lo_process();

		return
		(
			self.buffer_rise.mid() - rise_result,
			self.buffer_fall.mid() - fall_result,
		);
	}
}

struct CorrelOutput
{
	rolling_max_decay: f32,
	squelch: f32,

	buffer_rise: posi::Buffer<f32>,
	rise_sum: f32,
	rise_maxima: Vec<f32>,
	rise_max_rolling: f32,

	buffer_fall: posi::Buffer<f32>,
	fall_sum: f32,
	fall_maxima: Vec<f32>,
	fall_max_rolling: f32,
}

impl CorrelOutput
{
	fn new
	(
		size: usize,
	) ->
		CorrelOutput
	{
		CorrelOutput
		{
			rolling_max_decay: 1.0 - (0.1 / size as f32),
			squelch: 0.0,

			buffer_rise: posi::Buffer::new(size),
			rise_sum: 0.0,
			rise_maxima: Vec::new(),
			rise_max_rolling: 0.0,

			buffer_fall: posi::Buffer::new(size),
			fall_sum: 0.0,
			fall_maxima: Vec::new(),
			fall_max_rolling: 0.0,
		}
	}

	fn ingest
	(
		&mut self,
		new_rise: f32,
		new_fall: f32,
	) {
		let old_rise = self.buffer_rise.last();
		let old_fall = self.buffer_fall.last();

		self.rise_sum += new_rise.abs();
		self.rise_sum -= old_rise.abs();

		self.fall_sum += new_fall.abs();
		self.fall_sum -= old_fall.abs();
		
		if let Some(index) = self.rise_maxima.iter().position(|&x| x == old_rise)
		{
			self.rise_maxima.remove(index);
		}

		if let Some(index) = self.fall_maxima.iter().position(|&x| x == old_fall)
		{
			self.fall_maxima.remove(index);
		}

		if (
			self.rise_maxima.len() == 0
			||
			Some(&new_rise) >= self.rise_maxima.last()
		) {
			self.rise_maxima.push(new_rise);
		}

		if (
			self.fall_maxima.len() == 0
			||
			Some(&new_fall) >= self.fall_maxima.last()
		) {
			self.fall_maxima.push(new_fall);
		}

		self.buffer_rise.shift_in(new_rise);
		self.buffer_fall.shift_in(new_fall);

		let current_rise = self.buffer_rise.mid();
		let current_fall = self.buffer_fall.mid();

		self.rise_max_rolling *= self.rolling_max_decay; 
		self.fall_max_rolling *= self.rolling_max_decay; 

		self.rise_max_rolling = self.rise_max_rolling.max(current_rise).max(current_fall / 2.0);
		self.fall_max_rolling = self.fall_max_rolling.max(current_fall).max(current_rise / 2.0);
	}

	fn have_bit
	(
		&self,
	) ->
		Option<bool>
	{
		let current_rise = self.buffer_rise.mid();
		let current_fall = self.buffer_fall.mid();

		let rise_max = *self.rise_maxima.last().unwrap_or(&0.0);
		let fall_max = *self.fall_maxima.last().unwrap_or(&0.0);

		let mut have_rise:bool = false;
		let mut have_fall:bool = false;

		if (
			current_rise >= self.squelch
			&&
			current_rise >= rise_max
			&&
			current_rise >= fall_max * 7.0 / 8.0
			&&
			current_rise >= self.rise_max_rolling / 2.0
		) {
			have_rise = true;
		}

		if (
			current_fall >= self.squelch
			&&
			current_fall >= fall_max
			&&
			current_fall >= rise_max * 7.0 / 8.0
			&&
			current_fall >= self.fall_max_rolling / 2.0
		) {
			have_fall = true;
		}

		return match (have_rise, have_fall)
		{
			( true, false) => Some(true ),
			(false, true ) => Some(false),
			( true, true ) => Some(current_rise > current_fall),
			(false, false) => None,
		};
	}
}

struct ByteAccumulator
{
	byte: u8,
	bit: u8,
	since_bit: usize,
	abandon: usize,
}

impl ByteAccumulator
{
	fn new
	(
		abandon: usize,
	) ->
		ByteAccumulator
	{
		ByteAccumulator
		{
			byte: 0x00,
			bit: 0,
			since_bit: 0,
			abandon: abandon,
		}
	}

	fn accumulate
	(
		&mut self,
		possible_bit: Option<bool>,
	) ->
		Option<u8>
	{
		let mut output:Option<u8> = None;

		if let Some(new_bit) = possible_bit
		{
			self.since_bit = 0;
			self.byte |= (new_bit as u8).reverse_bits() >> self.bit;
			self.bit += 1;

			if self.bit >= 8
			{
				output = Some(self.byte);
				self.byte = 0x00;
				self.bit = 0;
			}
		}
		else
		{
			self.since_bit += 1;

			if self.since_bit > self.abandon
			{
				self.byte = 0x00;
				self.bit = 0;
			}
		}

		return output;
	}
}

pub struct Demodulator
{
	signal_input: SignalInput,
	correl_lopass: CorrelFilter,
	correl_hipass: CorrelFilter,
	correl_output: CorrelOutput,
	accumulator: ByteAccumulator,
}

impl Demodulator
{
	pub fn new
	(
		center_freq: f32,
		deviation: f32,
		symbol_length: f32,
	) -> 
		Demodulator
	{
		Demodulator
		{
			signal_input: SignalInput::new(center_freq, deviation, symbol_length),
			correl_lopass: CorrelFilter::new(deviation / 2.0),                       // these filter frequencies may need adjustment.
			correl_hipass: CorrelFilter::new(deviation / 4.0),                       // we don't really know how to set them Correctly, but they seem fairly easy to game.
			correl_output: CorrelOutput::new(symbol_length.round() as usize),
			accumulator: ByteAccumulator::new((symbol_length * 17.0 / 16.0).round() as usize),
		}
	}

	pub fn process
	(
		&mut self,
		new_sample: f32,
	) -> 
		Option<u8>
	{
		self.signal_input.ingest(new_sample);
		let (correl_rise, correl_fall) = self.signal_input.process();

		self.correl_lopass.ingest(correl_rise.abs(), correl_fall.abs());
		let (rise_lo, fall_lo) = self.correl_lopass.lo_process();

		self.correl_hipass.ingest(rise_lo, fall_lo);
		let (rise_hi, fall_hi) = self.correl_hipass.hi_process();

		self.correl_output.ingest(rise_hi, fall_hi);
		let possible_bit = self.correl_output.have_bit();

		let possible_byte = self.accumulator.accumulate(possible_bit);

		return possible_byte;
	}

	pub fn set_squelch
	(
		&mut self,
		value: f32,
	) {
		self.correl_output.squelch = value;
	}

	pub fn dump_correl
	(
		&self,
	) -> 
		(f32, f32)
	{
		(
			//self.correl_output.buffer_rise.mid(),
			//self.correl_output.buffer_fall.mid(),
			self.correl_output.rise_max_rolling,
			self.correl_output.fall_max_rolling,
		)
	}
}
